#!/usr/bin/python
# Copyright 2016 Mellanox Technologies, Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sys
import yaml
import logging
import re

MLNX_SECTION = 'mellanox-rhel-osp'
HIERA_CONFIG_FILE = '/etc/puppet/hiera.yaml'
MLNX_RHEL_OSP_FILE = '/opt/mlnx_scripts/mellanox-rhel-osp.yaml'
EXTRA_CONFIG_FILE = '/etc/puppet/hieradata/extraconfig.json'
MLNX_DRIVERS_LIST = {'ConnectX-3': {'eth_driver': 'mlx4_en',
                                    'ib_driver': 'eth_ipoib'},
                     'ConnectX-4': {'eth_driver': 'mlx5_core',
                                    'ib_driver': 'eth_ipoib'}}

MLNX_DRIVERS = set([MLNX_DRIVERS_LIST[card][net]
                    for card in MLNX_DRIVERS_LIST
                    for net in MLNX_DRIVERS_LIST[card]])
ETH_DRIVERS = set([MLNX_DRIVERS_LIST[card][net]
                   for card in MLNX_DRIVERS_LIST
                   for net in MLNX_DRIVERS_LIST[card]
                   if net == 'eth_driver'])
IB_DRIVERS = MLNX_DRIVERS - ETH_DRIVERS
LOG_FILE = '/var/log/mellanox-rhel-osp.log'


class MellanoxSettingsException(Exception):
    pass


class MellanoxSettings(object):

    data = None
    mlnx_data = dict()

    @classmethod
    def obtain_mlnx_data(cls):
        hiera_cmd = 'hiera -c {0} neutron_sriov_agent_enabled'.\
            format(HIERA_CONFIG_FILE)
        sriov_en = os.popen(hiera_cmd).readlines()
        if sriov_en[0].strip() == 'true':
            cls.mlnx_data['sriov_en'] = True
        else:
            cls.mlnx_data['sriov_en'] = False
        if cls.mlnx_data['sriov_en'] is True:
            hiera_cmd = 'hiera -c {0} tripleo::host::sriov::number_of_vfs'.\
                format(HIERA_CONFIG_FILE)
            interface_vfs = os.popen(hiera_cmd).readlines()
            for int_vfs in interface_vfs:
                interface = re.findall('\["(.*)\:', int_vfs)
                num_vfs = re.findall('\:(.*)\"]', int_vfs)
            cls.mlnx_data['sriov_interface'] = interface[0]
            cls.mlnx_data['num_of_vfs'] = int(num_vfs[0])

    @classmethod
    def get_card_type(cls, driver):
        for card in MLNX_DRIVERS_LIST.keys():
            if driver in MLNX_DRIVERS_LIST[card].values():
                return card

    @classmethod
    def add_cx_card(cls):
        current_driver = cls.mlnx_data['driver']
        if current_driver in ETH_DRIVERS:
            cls.mlnx_data['network_type'] = 'ethernet'
            cls.mlnx_data['cx_card'] = cls.get_card_type(current_driver)

        network_info_msg = 'Detected Network Type is' + \
                           ': {0} '.format(cls.mlnx_data['network_type'])
        card_info_msg = 'Detected Card Type is: {0} '.format(cls.mlnx_data[
                                                             'cx_card'])
        logging.info(network_info_msg)
        logging.info(card_info_msg)

    @classmethod
    def add_driver(cls):
        cmd_ethtool = "ethtool -i {0}".format(cls.mlnx_data['sriov_interface'])
        cmd_ethtool_res = os.popen(cmd_ethtool).readlines()
        for line in cmd_ethtool_res:
            if 'driver:' in line:
                driver = line.strip('driver:').strip()
                cls.mlnx_data['driver'] = driver

    @classmethod
    def add_reboot_required(cls):
        cls.mlnx_data['reboot_required'] = False

        # Check if mstflint is installed or not, if not, then install it
        cmd_mstflint_installed = 'rpm -qa | grep mstflint'
        mstflint_installed = os.popen(cmd_mstflint_installed).readlines()
        if 'mstflint' not in mstflint_installed:
            logging.info('Installing mstflint package')
            cmd_install_mstflint = 'yum install mstflint -y'
            os.popen(cmd_install_mstflint)

        # Setting reboot_required to true if burned number of
        # vfs != user_num_vfs OR SRIOV_EN is false and sriov
        # is true in user passed params
        cmd_interface_pci = 'ethtool -i {0} | grep bus-info '\
                            '| awk {{\'print$2\'}}'.format(
                                cls.mlnx_data['sriov_interface'])
        cls.mlnx_data['interface_pci'] = os.popen(cmd_interface_pci).read().\
            strip()
        mstconfig = os.popen('mstconfig -d {0} q'.format(
            cls.mlnx_data['interface_pci'])).readlines()
        for line in mstconfig:

            # Take just the number from num of vfs
            if 'NUM_OF_VFS' in line:
                num_of_vfs = ''.join(re.findall('([0-9]+)', line)).strip()
                if num_of_vfs and \
                    int(num_of_vfs) != cls.mlnx_data['num_of_vfs'] \
                        and cls.mlnx_data['sriov_en']:
                    cls.mlnx_data['reboot_required'] = True
                    logging.info(
                        'reboot_required is true as {0} is != {1}'.format(
                            num_of_vfs, cls.mlnx_data['num_of_vfs']))
                    return

            if 'SRIOV_EN' in line:
                sriov_en = line
                if cls.mlnx_data['sriov_en'] and "False" in sriov_en:
                    cls.mlnx_data['reboot_required'] = True
                    logging.info(
                        'reboot_required is true as SRIOV_EN is False'
                        ' and SRIOV is true')

    @classmethod
    def burn_FW(cls):
        if cls.mlnx_data['reboot_required'] is True:
            os.popen('mstconfig -y -d {0} s SRIOV_EN=1 NUM_OF_VFS={1}'
                     .format(cls.mlnx_data['interface_pci'],
                             cls.mlnx_data['num_of_vfs']))

    @classmethod
    def update_role_settings(cls):
        cls.mlnx_data['reboot_required'] = False
        node_id = os.popen('hiera -c {0} bootstrap_nodeid'.
                           format(HIERA_CONFIG_FILE)).readlines()
        if 'compute' in node_id[0].strip():
            cls.obtain_mlnx_data()
            if cls.mlnx_data['sriov_en'] is True:
                # realize the driver in use (eth/ib)
                cls.add_driver()

                # detect ConnectX card
                cls.add_cx_card()

                # add reboot required condition
                cls.add_reboot_required()
                cls.burn_FW()

    @classmethod
    def read_from_yaml(cls, settings_file):
        try:
            fd = open(settings_file, 'r')
        except IOError:
            raise MellanoxSettingsException("Given YAML file {0} doesn't "
                                            "exist".format(settings_file))
        try:
            data = yaml.load(fd)
        except yaml.YAMLError as exc:
            if hasattr(exc, 'problem_mark'):
                mark = exc.problem_mark
                raise MellanoxSettingsException(
                    "Faild parsing YAML file {0}: error position "
                    "({2},{3})".format(mark.line + 1, mark.column + 1)
                )
        finally:
            fd.close()
        cls.data = data
        cls.mlnx_data = data[MLNX_SECTION]

    @classmethod
    def write_to_yaml(cls, mlnx_settings_file):

        # choose only the edited sections
        data = {}
        data = cls.data
        data[MLNX_SECTION] = cls.mlnx_data

        try:
            fd = open(mlnx_settings_file, 'w')
            yaml.dump(data, fd, default_flow_style=False)

        except IOError:
            error_msg = "Failed writing changes to {0}".format(
                mlnx_settings_file)
            raise MellanoxSettingsException(error_msg)
        finally:
            if fd:
                fd.close()

    @classmethod
    def update_settings(cls):
        try:
            cls.read_from_yaml(EXTRA_CONFIG_FILE)
            cls.update_role_settings()
            cls.write_to_yaml(MLNX_RHEL_OSP_FILE)
        except MellanoxSettingsException as exc:
            error_msg = "Couldn't add Mellanox settings to " \
                "{0}: {1}\n".format(MLNX_RHEL_OSP_FILE, exc)
            sys.stderr.write(error_msg)
            logging.error(error_msg)
            raise MellanoxSettingsException("Failed updating one or more "
                                            "setting files")


def main():
    logging.basicConfig(format='%(asctime)s %(filename)s %(message)s',
                        level=logging.DEBUG, filename=LOG_FILE)

    try:
        settings = MellanoxSettings()
        settings.update_settings()
    except MellanoxSettingsException as exc:
        error_msg = "Failed adding Mellanox settings: {0}\n".format(exc)
        sys.stderr.write(error_msg)
        logging.error(exc)
        sys.exit(1)
    success_msg = "Done adding Mellanox settings\n"
    sys.stdout.write(success_msg)
    logging.info(success_msg)
    sys.exit(0)


if __name__ == '__main__':
    main()
