#!/usr/bin/env python
# Copyright 2016 Mellanox Technologies, Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sys
import logging
import argparse
import textwrap
from collections import OrderedDict
from utils import run_command
from utils import load_file
from utils import check_bool
from utils import install_libguestsfs
from utils import check_libvirtd
from params import EnvParams
from params import OvercloudParams


parser = argparse.ArgumentParser(
    prog=os.path.basename(sys.argv[0]),
    formatter_class=argparse.RawDescriptionHelpFormatter,
    description=textwrap.dedent('''\
             Log file: {0}

             '''.format(EnvParams.LOG_FILE_PATH)))

parser.add_argument(
    "--iso_ofed_path",
    help='path to the OFED iso file',
    type=str,
    required=False)
parser.add_argument(
    "--overcloud_images_path",
    help='path to the overcloud images',
    type=str,
    required=True)
parser.add_argument(
    "--root_password",
    help='set root password for the overcloud nodes',
    type=str,
    required=True)
parser.add_argument(
    "--interfaces_drivers",
    help=argparse.SUPPRESS,
    type=str,
    required=False)
parser.add_argument(
    "--ovs_dpdk_rpms_folder_path",
    help='Path to ovs-dpdk rpms folder',
    type=str,
    required=False)

parser.add_argument(
    "--ceph_rdma_repo_folder_path",
    help='Path to CEPH RDMA repo folder, '
         'required when using ceph rdma storage node',
    type=str,
    required=False)

args = parser.parse_args()
ovc_image = os.path.join(args.overcloud_images_path,
                         EnvParams.OVC_FULL_IMAGE_NAME)


def load_overcloud_conf():
    """
    Load overcloud configuartions.
    """
    config = load_file(EnvParams.CONF_FILE_PATH)
    config['MofedInstallation'] = check_bool(config['MofedInstallation'])
    config['Dpdk']['dpdk_enable'] = check_bool(config['Dpdk']['dpdk_enable'])
    return config


def upload_OFED():
    """
    Upload OFED iso to the overcloud-full image.
    """
    ofed_d = extract_ofed_info()
    if 'os_version' in ofed_d:
        check_ovc_target_OS(ofed_d['os_version'].strip('rhel'))

    run_command('virt-customize -a {0} --mkdir {1}'
                .format(ovc_image,
                        OvercloudParams.OVC_OFED_DIR_PATH))

    run_command('virt-customize -a {0} --upload \
                {1}:{2}'.format(ovc_image,
                                args.iso_ofed_path,
                                OvercloudParams.OVC_MLNX_PATH))
    rename_ofed_iso(ofed_d)


def get_mlnx_OVS_DPDK_rpms():
    """
    Get mellanox OVS-DPDK rpms from bgate
    """
    run_command('rm -rf {0}'
                .format(OvercloudParams.TMP_OVS_DPDK_PACKAGES_SOURCE_DIR))
    run_command('mkdir -p {0}'
                .format(OvercloudParams.TMP_OVS_DPDK_PACKAGES_SOURCE_DIR))
    if args.ovs_dpdk_rpms_folder_path is None:
        for rpm in OvercloudParams.MLNX_OVS_DPDK_RPMS_URLS_LIST:
            run_command('wget {0} -P {1}'
                        .format(rpm,
                                OvercloudParams.
                                TMP_OVS_DPDK_PACKAGES_SOURCE_DIR))
    else:
        run_command('cp -r  {0}/* {1}'
                    .format(args.ovs_dpdk_rpms_folder_path,
                            OvercloudParams.TMP_OVS_DPDK_PACKAGES_SOURCE_DIR))


def upload_mlnx_OVS_DPDK_rpms():
    """
    Upload mellanox OVS-DPDK rpms to overcloud image
    """
    run_command('virt-customize -a {0} --copy-in {1}:{2}'
                .format(ovc_image,
                        OvercloudParams.TMP_OVS_DPDK_PACKAGES_SOURCE_DIR,
                        OvercloudParams.OVC_MLNX_PATH))
    """
    Remove the tmp directory
    """
    run_command('rm -rf {0}'
                .format(OvercloudParams.TMP_OVS_DPDK_PACKAGES_SOURCE_DIR))


def extract_ofed_info():
    """
    Extract ofed version, distro and arch from ofed files.
    """
    ofed_d = OrderedDict()
    run_command('sudo mkdir -p {0}'.format(EnvParams.MNT_OFED_DIR_PATH))
    run_command('sudo mount -o loop {0} {1}'
                .format(args.iso_ofed_path, EnvParams.MNT_OFED_DIR_PATH))

    if os.path.isfile(EnvParams.OFED_VERSION_PATH):
        ofed_d['ofed_version'] = open(EnvParams.OFED_VERSION_PATH) \
            .read().strip()
    if os.path.isfile(EnvParams.OFED_DISTRO_PATH):
        ofed_d['os_version'] = open(EnvParams.OFED_DISTRO_PATH).read().strip()
    if os.path.isfile(EnvParams.OFED_ARCH_PATH):
        ofed_d['arch'] = open(EnvParams.OFED_ARCH_PATH).read().strip()

    run_command('sudo umount {0}'.format(EnvParams.MNT_OFED_DIR_PATH))
    run_command('sudo rm -rf {0}'.format(EnvParams.MNT_OFED_DIR_PATH))
    return ofed_d


def check_ovc_target_OS(ofed_distro):
    """
    Check OFED OS version vs. overcloud target OS version.
    """
    run_command(
        "virt-customize -a {0} --run-command \"cat /etc/redhat-release "
        "| grep -oP \"[0-9]+\.[0-9]+\" | head -1 > /{1}\""
        .format(ovc_image,
                os.path.join(OvercloudParams.OVC_TARGET_OS_DIR_PATH,
                             OvercloudParams.OVC_TARGET_OS)))

    run_command('virt-copy-out -a {0} {1} {2}'
                .format(ovc_image,
                        os.path.join(OvercloudParams.OVC_TARGET_OS_DIR_PATH,
                                     OvercloudParams.OVC_TARGET_OS),
                        EnvParams.TMP_FILES_DIR_PATH))

    if ofed_distro != open(os.path.join(EnvParams.TMP_FILES_DIR_PATH,
                                        OvercloudParams.OVC_TARGET_OS)) \
            .read().strip():
        run_command("virt-customize -a {0} --run-command \
                   'cat /etc/redhat-release > {1}' "
                    .format(ovc_image,
                            OvercloudParams.OVC_TARGET_OS))

        run_command('virt-copy-out -a {0} {1} {2}'
                    .format(ovc_image,
                            os.path.join(
                                OvercloudParams.OVC_TARGET_OS_DIR_PATH,
                                OvercloudParams.OVC_TARGET_OS),
                            EnvParams.TMP_FILES_DIR_PATH))
        target_os = open(os.path.join(EnvParams.TMP_FILES_DIR_PATH,
                                      OvercloudParams.OVC_TARGET_OS)).read()
        if ofed_distro not in target_os:
            error_msg = 'Please use compatible OFED version for {0}\n' \
                        .format(target_os)
            logging.error(error_msg)
            sys.stderr.write(error_msg)
            sys.exit(1)


def rename_ofed_iso(ofed_d):
    """
    Rename ofed iso on overcloud if needed.
    """
    ofed_file = os.path.basename(os.path.normpath(args.iso_ofed_path))

    ofed_new_name = OvercloudParams.OFED_PREFIX
    for key, value in ofed_d.items():
        ofed_new_name = '{0}-{1}'.format(ofed_new_name, value)
    ofed_new_name = '{0}.iso'.format(ofed_new_name)

    if ofed_file != ofed_new_name:
        run_command("virt-customize -a {0} --run-command \
                   'mv {1} {2}' "
                    .format(ovc_image,
                            os.path.join(OvercloudParams.OVC_MLNX_PATH,
                                         ofed_file),
                            os.path.join(OvercloudParams.OVC_MLNX_PATH,
                                         ofed_new_name)))


def set_root_password():
    """
    Set root password.
    """
    run_command('virt-customize -a {0} --root-password password:{1}'
                .format(ovc_image,
                        args.root_password))


def upload_ovc_scripts():
    """
    Upload common and restart_service files to overcloud-full image.
    """
    run_command('virt-customize -a {0} --mkdir {1}'
                .format(ovc_image,
                        OvercloudParams.OVC_MLNX_SCRIPTS_DIR_PATH))

    run_command(
        'export LIBGUESTFS_BACKEND=direct && virt-customize -a {0} \
         --upload {1}:{2}'
        .format(ovc_image,
                EnvParams.COMMON_SCRIPT_PATH,
                OvercloudParams.OVC_MLNX_SCRIPTS_DIR_PATH))
    run_command(
        'virt-customize -a {0} \
         --upload {1}:{2}'
        .format(ovc_image,
                EnvParams.RESTART_SERVICE_PATH,
                OvercloudParams.OVC_MLNX_SCRIPTS_DIR_PATH))


def add_future_yum(repo_url, repo_name):
    """
    Add repo and disable repo.
    """
    # Add repository to overcloud-full image.
    run_command("virt-customize -a {0} --run-command \
                'yum-config-manager --add-repo {1}'"
                .format(ovc_image,
                        repo_url))

    # Disable the added repository.
    run_command("virt-customize -a {0} --run-command \
                'yum-config-manager --disable {1}'"
                .format(ovc_image,
                        repo_name))


def firstboot_install(repo_name, package):
    """
    Generate firstboot command to install packages from specific repo.
    """

    # Install package from the added repo.
    cmd = "yum install -y --nogpgcheck {0} --enablerepo={1}\n" \
          .format(package,
                  repo_name)

    return cmd


def update_overcloud_images():
    """
    Upload and update existing overcloud images.
    """
    run_command('source {0} && cd {1} && openstack overcloud image upload --image-path {1} \
                --update-existing'.format(os.path.join('~', 'stackrc'),
                                          args.overcloud_images_path))


def check_valid_configuration(config):
    """
    Check if iso_ofed_path is provided in case of MofedInstallation is True
    """
    if config['MofedInstallation'] is True:
        if args.iso_ofed_path is None:
            error_msg = 'Overcloud images prepation failed\n'\
                        'MOFED installation is enabled but there\'s no path '\
                        'passed for it\n'\
                        '--iso_ofed_path ISO_OFED_PATH\n'
            logging.error(error_msg)
            sys.stderr.write(error_msg)
            exit(1)
    if (config['Storage']['backend'] == 'ceph' and
            check_bool(config['Storage']['rdma_enable'])):
        if args.ceph_rdma_repo_folder_path is None:
            error_msg = 'Overcloud images prepation failed\n'\
                        'Ceph RDMA is enabled but there\'s no path '\
                        'passed for ceph rpms\n'\
                        '--ceph_rdma_repo_folder_path '\
                        'CEPH_RDMA_REPO_FOLDER_PATH\n'
            logging.error(error_msg)
            sys.stderr.write(error_msg)
            exit(1)


def upload_ceph_rdma_packages():
    """
    Upload CEPH RDMA packages to  overcloud images
    and create repo in /etc/yum.repos.d/
    """
    # Copy ceph rpms to temp directory
    run_command('rm -rf {0}'
                .format(OvercloudParams.TMP_CEPH_PACKAGES_SOURCE_DIR))
    run_command('mkdir -p {0}'
                .format(OvercloudParams.TMP_CEPH_PACKAGES_SOURCE_DIR))
    run_command('cp -r  {0}/* {1}'
                .format(args.ceph_rdma_repo_folder_path,
                        OvercloudParams.TMP_CEPH_PACKAGES_SOURCE_DIR))

    # Upload ceph rpms to overcloud image
    run_command('virt-customize -a {0} --copy-in {1}:{2}'
                .format(ovc_image,
                        OvercloudParams.TMP_CEPH_PACKAGES_SOURCE_DIR,
                        OvercloudParams.OVC_MLNX_PATH))

    # Remove the tmp directory
    run_command('rm -rf {0}'
                .format(OvercloudParams.TMP_CEPH_PACKAGES_SOURCE_DIR))

    # Create repo in /etc/yum.repos.d/
    run_command('virt-copy-in -a {0} {1} {2}'.format(
        ovc_image,
        EnvParams.CEPH_RDMA_REPO_FILE,
        OvercloudParams.YUM_REPO_FOLDER))


def prepare_ovc_image():

    logging.basicConfig(format='%(asctime)s  %(filename)s %(message)s',
                        level=logging.DEBUG, filename=EnvParams.LOG_FILE_PATH)

    run_command('mkdir -p {0}'.format(EnvParams.TMP_FILES_DIR_PATH))
    config = load_overcloud_conf()
    check_valid_configuration(config)
    install_libguestsfs()
    check_libvirtd()
    if config['MofedInstallation'] is True:
        upload_OFED()
    if config['Dpdk']['dpdk_enable'] is True:
        get_mlnx_OVS_DPDK_rpms()
        upload_mlnx_OVS_DPDK_rpms()
    set_root_password()
    upload_ovc_scripts()

    if (config['Storage']['backend'] == 'ceph' and
            check_bool(config['Storage']['rdma_enable'])):
        upload_ceph_rdma_packages()

    update_overcloud_images()
    output = 'Overcloud images prepared successfully, run introspection again'
    logging.info(output)
    sys.stderr.write(output + '\n')
    sys.exit(0)


if __name__ == '__main__':
    prepare_ovc_image()
